/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : SCPCrossoverFusion.cc                                    */
/*                                                                       */
/*       Date : March 5, 1998                                            */
/* Updated at : March 5, 1998                                            */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/**
    Beasley&Chu
**/
/*************************************************************************/

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

#include <Chromosomes.h>
#include <ChromosomesBag.h>
#include <WeightedRandomPickerSet.h>
#include <tSet.h>

#include "SCPCrossoverFusion.h"
#include "SCPChromosomes.h"
#include "SetCoverProblem.h"

/*************************************************************************/
/*************************************************************************/

SCPCrossoverFusion::SCPCrossoverFusion() { debugOff(); }

/*************************************************************************/
/*************************************************************************/

SCPCrossoverFusion::~SCPCrossoverFusion()
{  cout << " de allocate SCPCrossoverFusion " << endl; }

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossoverFusion::applyOLD(Chromosomes* par1, Chromosomes * par2)
{   
    if(debug) cout << " I am in SCPCrossoverFusion::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();
    Column clm(numOfRows);
    Column cpr1(numOfCols);
    Column cpr2(numOfCols);
    Column coff(numOfCols);
    Column d1_2(numOfRows);
    Column d2_1(numOfRows);
    cost_t cost1 = parent1->getCost();
    cost_t cost2 = parent2->getCost();
    int i;

    if(debug) cout << " 1. I am in SCPCrossoverFusion::apply() " << endl;

    SCPChromosomes *offspring = new SCPChromosomes(parent1->getProblem());

    if(debug) cout << " 2. I am in SCPCrossoverFusion::apply() " << endl;

    /** Find the probability of each gene */
    double p1 = 1.0 * (cost1)/(cost1+cost2);
    double p2 = 1.0 * (cost2)/(cost1+cost2);

    tSet<int> cl = parent1->getColumns();
    if(debug) cout << " cl : " << cl << endl;

    int clIndx;
    for(i=0;i<cl.getSize();i++) 
	if( cl.getElementAtSafe(i,clIndx) ) cpr1.set(clIndx);

    cl = parent2->getColumns();
    for(i=0;i<cl.getSize();i++) 
	if( cl.getElementAtSafe(i,clIndx) ) cpr2.set(clIndx);

    for(i=0;i<numOfCols;i++) 
	if( cpr1.test(i)!=cpr2.test(i) )
	{
	  float r = 0.01 * (rand()%101);
	  if(r<=p1) 
	  { 
	     if(cpr1.test(i)) coff.set(i);
	     else coff.reset(i);
	  }
	  else 
	  { 
	     if(cpr2.test(i)) coff.set(i);
	     else coff.reset(i);
	  }
	}
	else { if( cpr1.test(i) ) coff.set(i);
	       else coff.reset(i); 
  	     }

    if(debug)
    cout << " cpr1: " << cpr1 << endl
         << " cpr1: " << cpr2 << endl
         << " coff: " << coff << endl;

    for(i=0;i<numOfCols;i++) if( coff.test(i) ) offspring->addColumn(i);

    if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring : " << *offspring << endl;

    }// end of if

    offspring->restoreBack();
    offspring->reduce();

    if( debug )
    {
       ofstream ou1("/tmp/fusion.out",ios::app);
       ou1  << " cpr1: " << cpr1 << endl
            << " cpr1: " << cpr2 << endl
            << " coff: " << coff << endl;
       ou1  << " parent1   : " << *parent1 << endl;
       ou1  << " parent2   : " << *parent2 << endl;
       ou1  << " offspring : " << *offspring << endl;
       ou1.close();
    }// end of if

    ChromosomesBag ret(1);

    ret.addChromosomes(offspring);

    return ret;

}// end of apply()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossoverFusion::apply(Chromosomes* par1, Chromosomes * par2)
{   
    if(debug) cout << " I am in SCPCrossoverFusion::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();
    Column clm(numOfRows);
    Column cpr1(numOfCols);
    Column cpr2(numOfCols);
    Column coff(numOfCols);
    Column d1_2(numOfRows);
    Column d2_1(numOfRows);
    cost_t cost1 = parent1->getCost();
    cost_t cost2 = parent2->getCost();
    int i;

    if(debug) cout << " 1. I am in SCPCrossoverFusion::apply() " << endl;

    SCPChromosomes *offspring = new SCPChromosomes(parent1->getProblem());

    if(debug) cout << " 2. I am in SCPCrossoverFusion::apply() " << endl;

    /** Find the probability of each gene */
    double p1 = 1.0 * (cost1)/(cost1+cost2);
    double p2 = 1.0 * (cost2)/(cost1+cost2);

    tSet<int> pa1s = parent1->getColumns();
    tSet<int> pa2s = parent2->getColumns();
    tSet<int> p1_and_p2 = pa1s * pa2s;
    tSet<int> p1_p2 = pa1s - pa2s;
    tSet<int> p2_p1 = pa2s - pa1s;
    int   clIndx;
    float r;

    // Common values should be there
    for(i=0;i<p1_and_p2.getSize();i++) 
	if( p1_and_p2.getElementAtSafe(i,clIndx) ) offspring->addColumn(clIndx);

    // Look at the different bits between p1 and p2
    for(i=0;i<p1_p2.getSize();i++) 
       if( p1_p2.getElementAtSafe(i,clIndx) ) 
       {
	  r = 0.01 * (rand()%101);
	  if(r<=p1)  offspring->addColumn(clIndx);  
       }

    // Look at the different bits between p2 and p1
    for(i=0;i<p2_p1.getSize();i++) 
       if( p2_p1.getElementAtSafe(i,clIndx) ) 
       {
	  r = 0.01 * (rand()%101);
	  if(r>p1)  offspring->addColumn(clIndx);  
       }

    if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring : " << *offspring << endl;

    }// end of if

    offspring->restoreBack();
    offspring->reduce();

    if( debug )
    {
       ofstream ou1("/tmp/fusion.out",ios::app);
       ou1  << " cpr1: " << cpr1 << endl
            << " cpr1: " << cpr2 << endl
            << " coff: " << coff << endl;
       ou1  << " parent1   : " << *parent1 << endl;
       ou1  << " parent2   : " << *parent2 << endl;
       ou1  << " offspring : " << *offspring << endl;
       ou1.close();

    }// end of if

    ChromosomesBag ret(1);

    ret.addChromosomes(offspring);

    return ret;

}// end of apply()

/*************************************************************************/
/*************************************************************************/

void SCPCrossoverFusion::applyf( Chromosomes* par1, Chromosomes * par2,
				 SCPChromosomes* offspring)
{   
    //if(debug) cout << " I am in SCPCrossoverFusion::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();

    cost_t cost1 = parent1->getCost();
    cost_t cost2 = parent2->getCost();
    int i;

    //if(debug) cout << " 1. I am in SCPCrossoverFusion::apply() " << endl;

    /** Find the probability of each gene */
    double p1 = 1.0 * (cost1)/(cost1+cost2);
    double p2 = 1.0 * (cost2)/(cost1+cost2);

    tSet<int> pa1s = parent1->getColumns();
    tSet<int> pa2s = parent2->getColumns();
    tSet<int> p1_and_p2 = pa1s * pa2s;
    tSet<int> p1_p2 = pa1s - pa2s;
    tSet<int> p2_p1 = pa2s - pa1s;
    int   clIndx;
    float r;

    // Common values should be there
    for(i=0;i<p1_and_p2.getSize();i++) 
	if( p1_and_p2.getElementAtSafe(i,clIndx) ) offspring->addColumn(clIndx);

    // Look at the different bits between p1 and p2
    for(i=0;i<p1_p2.getSize();i++) 
       if( p1_p2.getElementAtSafe(i,clIndx) ) 
       {
	  r = 0.01 * (rand()%101);
	  if(r<=p1)  offspring->addColumn(clIndx);  
       }

    // Look at the different bits between p2 and p1
    for(i=0;i<p2_p1.getSize();i++) 
       if( p2_p1.getElementAtSafe(i,clIndx) ) 
       {
	  r = 0.01 * (rand()%101);
	  if(r>p1)  offspring->addColumn(clIndx);  
       }

    /*
    if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring : " << *offspring << endl;

    }// end of if
    */

    offspring->restoreBack();
    offspring->reduce();

    /*
    if( debug )
    {
       ofstream ou1("/tmp/fusion.out",ios::app);
       ou1  << " cpr1: " << cpr1 << endl
            << " cpr1: " << cpr2 << endl
            << " coff: " << coff << endl;
       ou1  << " parent1   : " << *parent1 << endl;
       ou1  << " parent2   : " << *parent2 << endl;
       ou1  << " offspring : " << *offspring << endl;
       ou1.close();

    }// end of if
    */

}// end of applyf()

/*************************************************************************/
/*************************************************************************/

SCPCrossoverFusion::apply( Chromosomes* par1, Chromosomes * par2,
	                   Chromosomes** offsprings)
{   
    //if(debug) cout << " I am in SCPCrossoverFusion::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SCPChromosomes *offspring = (SCPChromosomes *) offsprings[0];
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();

    cost_t cost1 = parent1->getCost();
    cost_t cost2 = parent2->getCost();
    int i;

    //if(debug) cout << " 1. I am in SCPCrossoverFusion::apply() " << endl;

    /** Find the probability of each gene */
    double p1 = 1.0 * (cost1)/(cost1+cost2);
    double p2 = 1.0 * (cost2)/(cost1+cost2);

    tSet<int> pa1s = parent1->getColumns();
    tSet<int> pa2s = parent2->getColumns();
    tSet<int> p1_and_p2 = pa1s * pa2s;
    tSet<int> p1_p2 = pa1s - pa2s;
    tSet<int> p2_p1 = pa2s - pa1s;
    int   clIndx;
    float r;

    // Common values should be there
    for(i=0;i<p1_and_p2.getSize();i++) 
	if( p1_and_p2.getElementAtSafe(i,clIndx) ) offspring->addColumn(clIndx);

    // Look at the different bits between p1 and p2
    for(i=0;i<p1_p2.getSize();i++) 
       if( p1_p2.getElementAtSafe(i,clIndx) ) 
       {
	  r = 0.01 * (rand()%101);
	  if(r<=p1)  offspring->addColumn(clIndx);  
       }

    // Look at the different bits between p2 and p1
    for(i=0;i<p2_p1.getSize();i++) 
       if( p2_p1.getElementAtSafe(i,clIndx) ) 
       {
	  r = 0.01 * (rand()%101);
	  if(r>p1)  offspring->addColumn(clIndx);  
       }

    /*
    if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring : " << *offspring << endl;

    }// end of if
    */

    offspring->restoreBack();
    offspring->reduce();

    /*
    if( debug )
    {
       ofstream ou1("/tmp/fusion.out",ios::app);
       ou1  << " cpr1: " << cpr1 << endl
            << " cpr1: " << cpr2 << endl
            << " coff: " << coff << endl;
       ou1  << " parent1   : " << *parent1 << endl;
       ou1  << " parent2   : " << *parent2 << endl;
       ou1  << " offspring : " << *offspring << endl;
       ou1.close();

    }// end of if
    */

    return 1;

}// end of apply()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossoverFusion::apply(Chromosomes* par1, Chromosomes * par2, float percent)
{   
    return apply(par1,par2);
}
